@@ -9,16 +9,33 @@ use Carp;
 use base 'Exporter';
 our @EXPORT = qw(patronize getProxies);
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
+
+BEGIN {
+    if (defined &CORE::read) {
+	*CORE::GLOBAL::read = sub (*\$$;$) {
+	    $Patro::read_sysread_flag = 'read';
+	    goto &CORE::read if defined &CORE::read;
+	};
+	*CORE::GLOBAL::sysread = sub (*\$$;$) {
+	    $Patro::read_sysread_flag = 'sysread';
+	    goto &CORE::sysread if defined &CORE::sysread;
+	};
+    } else {
+	$Patro::read_sysread_flag = 'read?';
+    }
+}
 
 sub import {
     my ($class, @args) = @_;
     my @tags = grep /^:/, @args;
     @args = grep !/^:/, @args;
+    $Patro::SECURE = 1;
     foreach my $tag (@tags) {
 	if ($tag eq ':test') {
 	    require Patro::Server;
 	    Patro::Server->TEST_MODE;
+
 	    # some tests will check if the remote object has changed
 	    # after being manipulated by the proxy. This can only
 	    # happen with a threaded server (or with certain objects
@@ -42,20 +59,23 @@ sub import {
 				} : $_ } @_)
 	    };
 	    push @EXPORT, 'ok_threaded', 'xjoin';
-	    eval q~END {
-	        if ($Patro::Server::threads_avail) {
-	            $_->detach for threads->list(threads::running);
-	        }
-	    }~;
+	}
+	if ($tag eq ':insecure') {
+	    $Patro::SECURE = 0;
 	}
     }
-    if (eval "use threads;1") {
-	require Patro::CODE::Shareable;
-	Patro::CODE::Shareable->import;
-	Patro::CODE::Shareable->export_to_level(1, 'Patro::CODE::Shareable',
-						    'share','shared_clone');
-	*Patro::Server::shared_clone = *shared_clone;
-	*Patro::Server::share = *share;
+
+    if (defined($ENV{PATRO_THREADS}) &&
+	!$ENV{PATRO_THREADS}) {
+	$INC{'threads.pm'} = 1;
+	*threads::tid = sub { 0 };
+    }		
+    eval "use threads;1";
+    eval "use threadsx::shared";
+    $Patro::Server::threads_avail = $threads::threads;
+    if (defined $ENV{PATRO_THREADS}) {
+	no warnings 'redefine';
+	$Patro::Server::threads_avail = $ENV{PATRO_THREADS};
     }
     Patro->export_to_level(1, 'Patro', @args, @EXPORT);
 }
@@ -159,12 +179,10 @@ sub new {
 
 sub getProxies {
     my $patro = shift;
-    if (CORE::ref($patro) eq 'HASH') {
-	# arg to getProxies is config hash, not Patro object
-	my $cfg = $patro;
-	$patro = Patro->new($cfg);
+    if (CORE::ref($patro) ne 'Patro') {
+	$patro = Patro->new($patro);
     }
-    return @{$patro->{objs}};
+    return wantarray ? @{$patro->{objs}} : $patro->{objs}[0];
 }
 
 
@@ -174,9 +192,11 @@ sub getProxies {
 
 Patro - proxy access to remote objects
 
+
 =head1 VERSION
 
-0.11
+0.12
+
 
 =head1 SYNOPSIS
 
@@ -263,14 +283,70 @@ remote object.
     sub prod { my $self = shift; my $z=1; $z*=$_ for @{$_[0]->{vals}}; $z }
 
     # host 2
-    use Patro;'
+    use Patro;
     my $proxy = getProxies('config');
     print $proxy->prod;      # calls Barfie::prod($obj) on host1, 2 * 5 => 10
     $proxy += 4;             # calls Barfie '+=' sub on host1
     print $proxy->prod;      # 6 * 9 => 54
 
+=item * Code references
+
+Patro supports sharing code references and data structures that contain
+code references (think dispatch tables). Proxies to these code references
+can invoke the code, which will then run on the server.
+
+    # host 1
+    use Patro;
+    my $foo = sub { $_[0] + 42 };
+    my $d = {
+        f1 => sub { $_[0] + $_[1] },
+        f2 => sub { $_[0] * $_[1] },
+        f3 => sub { int( $_[0] / ($_[1] || 1) ) },
+        g1 => sub { $_[0] += $_[1]; 18 },
+    };
+    patronize($foo,$d)->to_file('config');
+    ...
+
+    # host 2
+    use Patro;
+    my ($p_foo, $p_d) = getProxies('config');
+    print $p_foo->(17);        # "59"   (42+17)
+    print $p_d->{f1}->(7,33);  # "40"   (7+33)
+    print $p_d->{f3}->(33,7);  # "4"    int(33/7)
+    ($x,$y) = (5,6);
+    $p_d->{g1}->($x,$y);
+    print $x;                  # "11"   ($x:6 += 5)
+
+=item * filehandles
+
+Filehandles can also be shared through the Patro framework.
+
+    # host 1
+    use Patro;
+    open my $fh, '>', 'host1.log';
+    patronize($fh)->to_file('config');
+    ...
+
+    # host 2
+    use Patro;
+    my $ph = getProxies('config');
+    print $ph "A log message for the server\n";
+
+Calling C<open> through a proxy filehandle presents some security concerns.
+A client could read or write any file on the server host visible to the
+server's user id. Or worse, a client could open a pipe through the handle
+to run an arbitrary command on the server. C<open> and C<close> operations
+on proxy filehandles will not be allowed unless the process running the
+Patro server imports C<Patro> with the C<:insecure> tag.
+
+Certain operations are not currently supported through proxy filehandles:
+C<truncate PROXY_FH,LENGTH>, C<flock PROXY_FH,OPERATION>,
+C<fcntl PROXY_FH,FUNCTION,SCALAR>, C<stat PROXY_FH>, and C<-X PROXY_FH>.
+
+
 =back
 
+
 =head1 FUNCTIONS
 
 =head2 patronize
@@ -328,6 +404,7 @@ with the remote object server. The client object contains information
 about how to communicate with the server and other connection 
 configuration.
 
+
 =head1 PROXIES
 
 Proxy references, as returned by the L<"getProxies"> function above,
@@ -449,6 +526,57 @@ it to determine whether to use a forked server or a threaded server
 to provide proxy access to objects. If this variable is not set,
 C<Patro> will use threads if the L<threads> module can be loaded.
 
+
+=head1 LIMITATIONS
+
+Proxy filehandles use the C<tie> mechanism for filehandles, so
+C<sysopen>, C<truncate>, C<flock>, C<fcntl>, C<stat HANDLE>,
+and C<-X HANDLE> functions are not supported on proxy filehandles,
+as documented in L<perltie>.
+
+When the server uses forks (because threads are unavailable or
+because L<"PATRO_THREADS"> was set to a false value), it is less
+practical to share variables between processes.
+When you manipulate a proxy reference, you are
+manipulating the copy of the reference running in a different process
+than the remote server. So you will not observe a change in the
+reference on the server (unless you use a class that does not save
+state in local memory, like L<Forks::Queue>).
+
+
+=head1 DOCUMENTATION AND SUPPORT
+
+Up-to-date (blead version) sources for C<Patro> are on github at
+L<https://github.com/mjob/p5-patro>
+
+You can find documentation for this module with the perldoc command.
+
+    perldoc Patro
+
+You can also look for information at:
+
+=over 4
+
+=item * RT: CPAN's request tracker
+
+Report bugs and request missing features at
+L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=Patro>
+
+=item * AnnoCPAN: Annotated CPAN documentation
+
+L<http://annocpan.org/dist/Patro>
+
+=item * CPAN Ratings
+
+L<http://cpanratings.perl.org/d/Patro>
+
+=item * Search CPAN
+
+L<http://search.cpan.org/dist/Patro/>
+
+=back
+
+
 =head1 LICENSE AND COPYRIGHT
 
 MIT License
