@@ -6,11 +6,18 @@ use Carp;
 use Storable;
 use MIME::Base64 ();
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
+
+my %proxyClasses = (
+    'Patro::N1' => 0,    # HASH
+    'Patro::N2' => 1,    # SCALAR
+    'Patro::N3' => 0,    # CODE
+    'Patro::N4' => 0,    # ARRAY
+    'Patro::N5' => 0 );  # GLOB
 
 sub isProxyRef {
     my ($pkg) = @_;
-    return $pkg eq 'Patro::N1' || $pkg eq 'Patro::N2' || $pkg eq 'Patro::N3';
+    return defined $proxyClasses{$pkg};
 }
 
 sub handle {
@@ -39,7 +46,7 @@ sub deserialize {
     }
 }
 
-# return a Patro::N1 or Patro::N2 object appropriate for the
+# return a Patro::Nx object appropriate for the
 # object metadata (containing id, ref, reftype values) and client.
 sub getproxy {
     my ($objdata,$client) = @_;
@@ -59,10 +66,10 @@ sub getproxy {
     }
 
     if ($proxy->{reftype} eq 'ARRAY') {
-	require Patro::N1;
+	require Patro::N4;
 	tie my @a, 'Patro::Tie::ARRAY', $proxy;
 	$proxy->{array} = \@a;
-	return bless \$proxy, 'Patro::N1';
+	return bless \$proxy, 'Patro::N4';
     }
 
     if ($proxy->{reftype} eq 'HASH') {
@@ -84,17 +91,27 @@ sub getproxy {
 		    args => [ @_ ],
 		    command => 'invoke',
 		    id => $proxy->{id}
-		} );
+		}, @_ );
 	};
 	return bless \$proxy, 'Patro::N3';
     }
 
+    if ($proxy->{reftype} eq 'GLOB') {
+	require Patro::N5;
+	require Symbol;
+	my $fh = Symbol::gensym();
+	tie *$fh, 'Patro::Tie::HANDLE', $proxy;
+	$proxy->{handle} = \*$fh;
+	return bless \$proxy, 'Patro::N5';
+    }
+
     croak "unsupported remote object reftype '$objdata->{reftype}'";
 }
 
 # make a request through a Patro::N's client, return the response
 sub proxy_request {
-    my ($proxy,$request) = @_;
+    my $proxy = shift;
+    my $request = shift;
     my $socket = $proxy->{socket};
     if (!defined $request->{context}) {
 	$request->{context} = defined(wantarray) ? 1 + wantarray : 0;
@@ -133,8 +150,34 @@ sub proxy_request {
 	croak $resp->{error};
     }
     if (exists $resp->{disconnect_ok}) {
+#	::xdiag("client: disconnect_ok received. Response is ",$resp);
 	return $resp;
     }
+
+    # before returning, handle side effects
+    if ($resp->{out} && ref($resp->{out}) eq 'ARRAY') {
+	for (my $i=0; $i<@{$resp->{out}}; ) {
+	    my $index = $resp->{out}[$i++];
+	    my $val = $resp->{out}[$i++];
+	    eval { $_[$index] = $val };
+	    if ($@) {
+		croak $@;
+	    }
+	}
+    }
+    if (defined $resp->{errno}) {
+	# the remote call set $!
+	$! = $resp->{errno};
+    }
+    if (defined $resp->{child_error}) {
+	# the remote call set $?
+	$? = $resp->{child_error};
+    }
+    if (defined $resp->{eval_error}) {
+	# the remote call set $@
+	$@ = $resp->{eval_error};
+    }
+
     if ($resp->{context} == 0) {
 	return;
     }
@@ -189,7 +232,7 @@ sub depatrol {
     return $obj;
 }
 
-# overload handling for Patro::N1 and Patro::N2
+# overload handling for Patro::N1, Patro::N2, and Patro::N4. N3 and N5 too?
 
 my %numeric_ops = map { $_ => 1 }
 qw# + - * / % ** << >> += -= *= /= %= **= <<= >>= <=> < <= > >= == != ^ ^=
@@ -250,6 +293,7 @@ sub overload_handler {
     return eval "\$str $op \$y";
 }
 
+sub extend_threads_shared { return }
 
 1;
 
