@@ -8,19 +8,13 @@ use Scalar::Util 'reftype';
 use POSIX ':sys_wait_h';
 require overload;
 
-our $threads_avail = eval "use threads; use threads::shared; 1";
-if (defined $ENV{PATRO_THREADS}) {
-    no warnings 'redefine';
-    $threads_avail = $ENV{PATRO_THREADS};
-    *threads::shared::tie::SPLICE = \&threads_shared_tie_SPLICE;
-}
-
+our $threads_avail;
 *sxdiag = sub {};
 if ($ENV{PATRO_SERVER_DEBUG}) {
     *sxdiag = *::xdiag;
 }
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 our @SERVERS :shared;
 our %OPTS = ( # XXX - needs documentation
     keep_alive => 30,
@@ -33,7 +27,7 @@ sub new {
     my $opts = shift;
 
     my $host = $ENV{HOSTNAME} // qx(hostname) // "localhost";
-    if (eval "require Sys::HostAddr;1") {
+    if ($INC{'Sys/HostAddr.pm'}) {
 	my $host2 = Sys::HostAddr->new->main_ip;
 	$host = $host2 if $host2;
     }
@@ -72,17 +66,14 @@ sub new {
 
     if ($threads_avail) {
 	for (@_) {
-	    if (CORE::ref($_) eq 'CODE') {
-		require Patro::CODE::Shareable;
-		Patro::CODE::Shareable->import;
-	    }
 	    local $threads::shared::clone_warn = undef;
-
-	    eval { $_ = shared_clone($_) };
-	    if ($@ =~ /CODE/) {
-		require Patro::CODE::Shareable;
+	    eval { $_ = threads::shared::shared_clone($_) };
+	    if ($@ =~ /CODE|GLOB/) {
+		require Patro::LeumJelly;
+		warn "I am really surprised to see this message";
+		warn $@;
 		$threads::shared::clone_warn = 0;
-		$_ = shared_clone($_);
+		$_ = threads::shared::shared_clone($_);
 	    }
 	}
     }
@@ -98,8 +89,10 @@ sub new {
 	$obj->{$num} = $o;
 	my $reftype = Scalar::Util::reftype($o);
 	my $ref = CORE::ref($o);
-	if ($ref eq 'Patro::CODE::Shareable') {
+	if ($ref eq 'threadsx::shared::code') {
 	    $ref = $reftype = 'CODE*';
+	} elsif ($ref eq 'threadsx::shared::glob') {
+	    $ref = $reftype = 'GLOB';
 	}
 	my $store = {
 	    ref => $ref,
@@ -107,7 +100,7 @@ sub new {
 	    id => $num
 	};
 	if (overload::Overloaded($o)) {
-	    if ($ref ne 'CODE' && $ref ne 'CODE*') {
+	    if ($ref ne 'CODE' && $ref ne 'CODE*' && $ref ne 'GLOB') {
 		$store->{overload} = _overloads($o);
 	    }
 	}
@@ -120,7 +113,14 @@ sub new {
     }, __PACKAGE__;
     $self->{config} = $self->config;
     $self->start_server;
-    push @SERVERS, $self;
+    eval { push @SERVERS, $self };
+    if (@SERVERS == 1) {
+	eval q~END {
+            if ($Patro::Server::threads_avail) {
+                $_->detach for threads->list(threads::running);
+	    }
+	}~;
+    }
     return $self;
 }
 
@@ -139,7 +139,7 @@ sub start_server {
 		    open my $fh, '>>', $self->{meta}{pid_file};
 		    flock $fh, 2;
 		    seek $fh, 0, 2;
-		    print $fh "$$-", threads->tid, "\n";
+		    print {$fh} "$$-", threads->tid, "\n";
 		    close $fh;
 		}
 		$self->accept_clients;
@@ -159,7 +159,7 @@ sub start_server {
 		open my $fh, '>>', $self->{meta}{pid_file};
 		flock $fh, 2;
 		seek $fh, 0, 2;
-		print $fh "$$\n";
+		print {$fh} "$$\n";
 		close $fh;
 	    }
 	    $self->accept_clients;
@@ -233,7 +233,7 @@ sub Patro::Config::to_file {
     if (!open($fh, '>', $file)) {
 	croak "Patro::Config::to_file: could not write cfga file '$file': $!";
     }
-    print $fh $self->to_string;
+    print {$fh} $self->to_string;
     close $fh;
     return $file;
 }
@@ -278,10 +278,10 @@ sub accept_clients {
 	my $server = $meta->{socket};
 	my $paddr = accept($client,$server);
 	if (!$paddr) {
-	    if ($!{EINTR} || $!{ECHILD}) {
-		next;
-	    }
-	    croak __PACKAGE__, ": accept $!";
+	    next if $!{EINTR};
+	    next if $!{ECHILD} || $!==10;   # !?! why $!{ECHILD} not suff on Lin?
+	    ::xdiag("accept failure, %errno is",\%!);
+	    croak __PACKAGE__, ": accept ", 0+$!," $!";
 	}
 	$meta->{last_connection} = time;
 
@@ -302,7 +302,7 @@ sub start_subserver {
 		open my $fh, '>>', $self->{meta}{pid_file};
 		flock $fh, 2;
 		seek $fh, 0, 2;
-		print $fh "$pid\n";
+		print {$fh} "$pid\n";
 		close $fh;
 	    }
 	    $self->{meta}{pids}{$pid}++;
@@ -321,7 +321,7 @@ sub start_subserver {
 	    open my $fh, '>>', $self->{meta}{pid_file};
 	    flock $fh, 2;
 	    seek $fh, 0, 2;
-	    print $fh "$$-", $subthread->tid, "\n";
+	    print {$fh} "$$-", $subthread->tid, "\n";
 	    close $fh;
 	}
 	$self->{meta}{pids}{"$$-" . $subthread->tid}++;
@@ -508,10 +508,13 @@ sub process_request {
 			      $obj,$command,$has_args,$args) };
 	return $@ ? $self->error_response($@) : $resp;
     }
-
+    
     elsif ($topic eq 'METHOD') {
 	my @r;
+	my @orig_args = $has_args ? @$args : ();
+	my @orig_refs = \ (@$args);  
 	my $obj = $self->{obj}{$id};
+	local $! = 0;
 	if ($ctx < 2) {
 	    @r = scalar eval { $has_args ? $obj->$command(@$args)
 				   : $obj->$command };
@@ -519,39 +522,97 @@ sub process_request {
 	    @r = eval { $has_args ? $obj->$command(@$args)
 			          : $obj->$command };
 	}
+	my $errno = 0 + $!;
 	if ($@) {
-	    return $self->error_response($@);
+	    return $self->error_response($@, bless{errno=>$errno},'.Patroclus');
+	}
+	my (@out,@outref);
+	for (my $i=0; $i<@orig_refs; $i++) {
+	    my $j = $i;
+	    if ($orig_refs[$i] eq \$args->[$j]) {
+		my $rt_orig = Scalar::Util::reftype($orig_args[$i]) // '';
+		my $rt = Scalar::Util::reftype($args->[$j]) // '';
+		if ('SCALAR' eq $rt_orig) {
+		    if ('SCALAR' eq $rt &&
+			${$orig_args[$i]} ne ${$args->[$j]}) {
+			push @outref, $i, ${$args->[$j]};
+		    } elsif (CORE::ref($args->[$j]) eq '') {
+			push @outref, $i, $args->[$j];
+		    }
+		} elsif ($orig_args[$i] ne $args->[$j]) {
+		    push @out, $i, $args->[$j];
+		}
+	    }
+	}
+	my @addl;
+	if (@out || @outref) {
+	    @addl = (bless { out => \@out, outref => \@outref }, '.Patroclus');
+	    if ($errno) {
+		$addl[0]{errno} = $errno;
+	    }
+	} elsif ($errno) {
+	    @addl = (bless { errno => $errno }, '.Patroclus');
 	}
 	if ($ctx >= 2) {
-	    return $self->list_response(@r);
+	    return $self->list_response(@r, @addl);
 	} elsif ($ctx == 1 && defined $r[0]) {
-	    return $self->scalar_response($r[0]);
+	    return $self->scalar_response($r[0], @addl);
 	} else {
-	    return $self->void_response;
+	    return $self->void_response(@addl);
 	}
     }
 
     elsif ($topic eq 'CODE') {
 	my $sub = $self->{obj}{$id};
 	my @r;
-	if (ref($sub) eq 'Patro::CODE::Shareable') {
+	if (ref($sub) eq 'threadsx::shared::code') {
 	    # not necessary if server uses perl >=v5.18
-	    $sub = $sub->_invoke;
+	    $sub = $sub->code;
 	}
+	my @orig_args = @$args;
+	local $! = 0;
 	if ($ctx < 2) {
 	    @r = scalar eval { $has_args ? $sub->(@$args) : $sub->() };
 	} else {
 	    @r = eval { $has_args ? $sub->(@$args) : $sub->() };
 	}
+	my $errno = 0 + $!;
 	if ($@) {
-	    return $self->error_response($@);
+	    if ($errno) {
+		$errno = bless { errno => $errno }, '.Patroclus';
+		return $self->error_response($@, $errno);
+	    } else {
+		return $self->error_response($@);
+	    }
+	}
+	my (@out,@outref);
+	for (my $i=0; $i<@orig_args; $i++) {
+	    if ('SCALAR' eq (Scalar::Util::reftype($orig_args[$i]) // '')) {
+		if ('SCALAR' eq (Scalar::Util::reftype($args->[$i]) // '') &&
+		    ${$orig_args[$i]} ne ${$args->[$i]}) {
+		    push @outref, $i, ${$args->[$i]};
+		} elsif (CORE::ref($args->[$i]) eq '') {
+		    push @outref, $i, $args->[$i];
+		}
+	    } elsif ($orig_args[$i] ne $args->[$i]) {
+		push @out, $i, $args->[$i];
+	    }
+	}
+	my @addl;
+	if (@out || @outref) {
+	    @addl = (bless { out => \@out, outref => \@outref }, '.Patroclus');
+	    if ($errno) {
+		$addl[0]{errno} = $errno;
+	    }
+	} elsif ($errno) {
+	    @addl = (bless { errno => $errno }, '.Patroclus');
 	}
 	if ($ctx >= 2) {
-	    return $self->list_response(@r);
+	    return $self->list_response(@r, @addl);
 	} elsif ($ctx == 1 && defined($r[0])) {
-	    return $self->scalar_response($r[0]);
+	    return $self->scalar_response($r[0], @addl);
 	} else {
-	    return $self->void_response;
+	    return $self->void_response(@addl);
 	}
     }
 
@@ -560,6 +621,12 @@ sub process_request {
 	return $self->process_request_OVERLOAD($obj,$command,$args,$ctx);
     }
 
+    elsif ($topic eq 'HANDLE') {
+	my $obj = $self->{obj}{$id};
+	return $self->process_request_HANDLE(
+	    $obj,$command,$ctx,$has_args,$args);
+    }
+
     else {
 	return $self->error_response(
 	    __PACKAGE__,": unrecognized topic '$topic'");
@@ -606,7 +673,7 @@ sub process_request_OVERLOAD {
 	return $self->error_response($@);
     }
     if ($threads_avail) {
-	$z = shared_clone($z);
+	$z = threads::shared::shared_clone($z);
     }
     return $self->scalar_response($z);
 }
@@ -694,10 +761,10 @@ sub process_request_ARRAY {
 	}	
 	return $self->list_response(@val);
     } elsif ($command eq 'PUSH') {
-	my $n = push @{$obj}, _share(@$args);
+	my $n = push @{$obj}, map threads::shared::shared_clone($_),@$args;
 	return $self->scalar_response($n);
     } elsif ($command eq 'UNSHIFT') {
-	my $n = unshift @$obj, _share(@$args);
+	my $n = unshift @$obj, map threads::shared::shared_clone($_),@$args;
 	return $self->scalar_response($n);
     } elsif ($command eq 'POP') {
 	return $self->scalar_response(pop @$obj);
@@ -710,22 +777,11 @@ sub process_request_ARRAY {
     die "tied ARRAY function '$command' not recognized";
 }
 
-sub _share {
-    if (!$threads_avail) {
-	return @_;
-    } else {
-	return map {
-	    CORE::ref($_) eq 'CODE' ? $_
-		: CORE::ref($_) ? shared_clone($_) : $_;
-	} @_
-    }
-}
-
 sub process_request_SCALAR {
     my ($self,$obj,$command,$has_args,$args) = @_;
     if ($command eq 'STORE') {
-	${$obj} = $args->[0];
-	return $self->scalar_response(_share(${$obj}));
+	${$obj} = threads::shared::shared_clone($args->[0]);
+	return $self->scalar_response(${$obj});
     } elsif ($command eq 'FETCH') {
 	my $return = $self->scalar_response(${$obj});
 	return $return;
@@ -733,6 +789,153 @@ sub process_request_SCALAR {
     die "topic 'SCALAR': command '$command' not recognized";
 }
 
+sub process_request_HANDLE {
+    my ($self,$obj,$command,$context,$has_args,$args) = @_;
+    my $fh = ref($obj) eq 'threadsx::shared::glob' ? $obj->glob : $obj;
+    if ($command eq 'PRINT') {
+	my $z = print {$fh} @$args;
+	return $self->scalar_response($z);
+    } elsif ($command eq 'PRINTF') {
+	if ($has_args) {
+	    my $template = shift @$args;
+	    my $z = printf {$fh} $template, @$args;
+	    return $self->scalar_response($z);
+	} else {
+	    # I don't think we can get here through the proxy
+	    my $z = printf {$fh} "";
+	    return $self->scalar_response($z);
+	}
+    } elsif ($command eq 'WRITE') {
+	if (@$args < 2) {
+	    return $self->error_response("Not enough arguments for syswrite");
+	}
+	return $self->scalar_response(syswrite $fh, $args->[0],
+				      $args->[1] // undef, $args->[2] // undef);
+    } elsif ($command eq 'READLINE') {
+	local $! = 0;
+	my @val;
+	if ($context > 1) {
+	    my @val = readline($fh);
+	    if ($!) {
+		push @val, bless { errno => 0+$! }, '.Patroclus';
+	    }
+	    return $self->list_response(@val);
+	} else {
+	    my $val = readline($fh);
+	    if ($!) {
+		return $self->scalar_response($val, {errno => 0+$!});
+	    } else {
+		return $self->scalar_response($val);
+	    }
+	}
+    } elsif ($command eq 'GETC') {
+	local $! = 0;
+	my $ch = getc($fh);
+	return $self->scalar_response($ch);
+    } elsif ($command eq 'READ' || $command eq 'READ?' ||
+	     $command eq 'SYSREAD') {
+	local $Patro::read_sysread_flag;  # don't clobber
+	if (@$args < 2) {
+	    # I don't think we can get here through the proxy
+	    return $self->error_response("Not enough arguments for read");
+	}
+	my ($buffer, $len, $off) = @$args;
+	my $z;
+	if ($command eq 'SYSREAD' || ($command eq 'READ?' && fileno($fh) >= 0)) {
+	    $z = sysread $fh, $buffer, $len, $off || 0;
+	} else {
+	    # sysread doesn't work, for example, on file handles opened
+	    # from a reference to a scalar
+	    $z = read $fh, $buffer, $len, $off || 0;
+	}
+	if ($@) {
+	    return $self->error_response($@);
+	} else {
+	    return $self->scalar_response($z, { out => [0,$buffer] })
+	}
+    } elsif ($command eq 'EOF') {
+	return $self->scalar_response( eof($fh) );
+    } elsif ($command eq 'FILENO') {
+	my $z = fileno($fh);
+	return $self->scalar_response($z);
+    } elsif ($command eq 'SEEK') {
+	if (@$args < 2) {
+	    return $self->error_response("Not enough arguments for seek");
+	} elsif (@$args > 2) {
+	    return $self->error_response("Too many arguments for seek");
+	} else {
+	    my $z = seek $fh, $args->[0], $args->[1];
+	    return $self->scalar_response($z);
+	}
+    } elsif ($command eq 'TELL') {
+	my $z = tell($fh);
+	return $self->scalar_response($z);
+    } elsif ($command eq 'BINMODE') {
+	local $! = 0;
+	my $z;
+	if (@$args) {
+	    $z = binmode $fh, $args->[0];
+	} else {
+	    $z = binmode $fh;
+	}
+	if (!$z) {
+	    return $self->scalar_response($z, { errno => 0+$! });
+	} else {
+	    return $self->scalar_response($z);
+	}
+    } elsif ($command eq 'CLOSE') {
+	if ($Patro::SECURE) {
+	    return $self->error_response(
+		"Patro: insecure CLOSE operation on proxy filehandle");
+	}
+	local $! = 0;
+	local $? = 0;
+	my $z = close $fh;
+	my $val = bless{ errno => 0 + $! }, '.Patroclus';
+	if ($?) { 
+	    $val->{child_error} = $?;
+	}
+	if ($val->{errno} || $val->{child_error}) {
+	    return $self->scalar_response($z, $val);
+	} else {
+	    return $self->scalar_response($z);
+	}
+    } elsif ($command eq 'OPEN') {
+	if ($Patro::SECURE) {
+	    return $self->error_response(
+		"Patro: insecure OPEN operation on proxy filehandle");
+	}
+	local $! = 0;
+	my $z;
+	my $mode = shift @$args;
+	if (@$args == 0) {
+	    $z = open $fh, $mode;
+	} else {
+	    my $expr = shift @$args;
+	    if (@$args == 0) {
+		$z = open $fh, $mode, $expr;
+	    } else {
+		$z = open $fh, $mode, $expr, @$args;
+	    }
+	}
+
+	# it is hard to set autoflush from the proxy.
+	# Since it is usually what you want, let's do it here.
+	if ($z) {
+	    my $fh_sel = select $fh;
+	    $| = 1;
+	    select $fh_sel;
+	}
+	
+	if ($!) {
+	    return $self->scalar_response($z, {errno => 0+$!});
+	} else {
+	    return $self->scalar_response($z);
+	}
+    }
+    return $self->error_response("tied HANDLE function '$command' not found");
+}
+    
 # we should not send any serialized references back to the client.
 # replace any references in the response with an
 # object id.
@@ -741,9 +944,14 @@ sub patrol {
     sxdiag("patrol: called on: ",defined($obj) ? "$obj" : "<undef>");
     return $obj unless ref($obj);
 
-    if ($threads_avail && ref($obj) eq 'CODE') {
-	$obj = Patro::CODE::Shareable->new($obj);
-	sxdiag("patrol: coderef converted");
+    if ($threads_avail) {
+	if (ref($obj) eq 'CODE') {
+	    $obj = threadsx::shared::code->new($obj);
+	    sxdiag("patrol: coderef converted");
+	} elsif (ref($obj) eq 'GLOB') {
+	    $obj = threadsx::shared::glob->new($obj);
+	    sxdiag("patrol: glob converted");
+	}
     }
 
     my $id = do {
@@ -755,9 +963,12 @@ sub patrol {
 	$self->{obj}{$id} = $obj;
 	my $ref = ref($obj);
 	my $reftype;
-	if ($ref eq 'Patro::CODE::Shareable') {
+	if ($ref eq 'threadsx::shared::code') {
 	    $ref = 'CODE';
 	    $reftype = 'CODE';
+	} elsif ($ref eq 'threadsx::shared::glob') {
+	    $ref = 'GLOB';
+	    $reftype = 'GLOB';
 	} else {
 	    $reftype = reftype($obj);
 	}
@@ -776,31 +987,53 @@ sub patrol {
 }
 
 sub void_response {
-    return +{ context => 0, response => undef };
+    my $addl = {};
+    if (@_ > 0 && CORE::ref($_[-1]) eq '.Patroclus') {
+	$addl = pop @_;
+    }
+    return +{ context => 0, response => undef, %$addl };
 }
 
 sub scalar_response {
-    my ($self,$val) = @_;
+    my ($self,$val,$addl) = @_;
     return +{
 	context => 1,
-	response => $val
+	response => $val,
+	$addl ? %$addl : ()
     };
 }
 
 sub list_response {
     my ($self,@val) = @_;
+    my $addl = {};
+    if (@val && CORE::ref($val[-1]) eq '.Patroclus') {
+	$addl = pop @val;
+    }
     return +{
 	context => 2,
-	response => \@val
+	response => \@val,
+	%$addl
     };
 }
 
 sub error_response {
     my ($self,@msg) = @_;
+    if (@msg && CORE::ref($msg[-1]) eq '.Patroclus') {
+	my $val = pop @msg;
+	return { error => join('',@msg), %$val };
+    }
     return { error => join('',@msg) };
 }
 
 sub TEST_MODE {
+    if ($INC{'perl5db.pl'}) {
+	::xdiag("TEST_MODE adjusted for debugging");
+	$OPTS{keep_alive} = 3600;
+	$OPTS{fincheck_freq} = 3500;
+	$OPTS{idle_timeout} = 3600;
+	alarm 9999;
+	return;
+    }
     $OPTS{keep_alive} = 2;
     $OPTS{fincheck_freq} = 2;
     $OPTS{idle_timeout} = 1;
@@ -809,64 +1042,6 @@ sub TEST_MODE {
     }
 }
 
-
-
-no warnings 'redefine';
-
-# core  threads::shared  has a limitation in that the  splice  function
-# can not be used on shared arrays. We can hijack  threads::shared::tie::SPLICE
-# with this function, that performs the splice operation without using
-# the splice function, to work around this limitation.
-# This is not a very efficient implementation, but it is better than
-# a sharp stick in the eye.
-#
-sub threads_shared_tie_SPLICE {
-    use Data::Dumper;
-    use B;
-
-    my ($tied,$off,$len,@list) = @_;
-    my @bav = B::AV::ARRAY($tied);
-    my $arraylen = 0 + @bav;
-    if ($off < 0) {
-	$off += $arraylen;
-	if ($off < 0) {
-	    croak "Modification of non-createable array value attempated, ",
-		"subscript $_[1]";
-	}
-    }
-    if (!defined $len || $len eq 'undef') {
-	$len = $arraylen - $off;
-    }
-    if ($len < 0) {
-	$len += $arraylen - $off;
-	if ($len < 0) {
-	    $len = 0;
-	}
-    }
-
-    my (@tmp, @val);
-    for (my $i=0; $i<$off; $i++) {
-	my $fetched = $bav[$i]->object_2svref;
-	push @tmp, $$fetched;
-    }
-    for (my $i=0; $i<$len; $i++) {
-	my $fetched = $bav[$i+$off]->object_2svref;
-	push @val, $$fetched;
-    }
-    push @tmp, @list;
-    for (my $i=$off+$len; $i<$arraylen; $i++) {
-	my $fetched = $bav[$i]->object_2svref;
-	push @tmp, $$fetched;
-    }
-    # is there a better way to clear the array?
-    $tied->STORESIZE($#tmp + 1);
-    $tied->POP for 0..$arraylen;
-
-    $tied->PUSH(@tmp);
-
-    return @val;
-}
-
 1;
 
 =head1 NAME
@@ -875,7 +1050,7 @@ Patro::Server - remote object server for Patro
 
 =head1 VERSION
 
-0.11
+0.12
 
 =head1 DESCRIPTION
 
